#include <stdio.h>
#include <CL/cl.h>
#include <string.h>
#include <time.h>
#include <stdlib.h>
#include <math.h>
#include <iostream>
#include <algorithm>
#include <cstdlib>
#include <vector>
#include <windows.h>
#include <CL/cl_ext.h>

// preprocessor defines
typedef float type;

//data size
#define DATA_SIZE 1<<25
#define NLOOPS 5000

LARGE_INTEGER frequency;
LARGE_INTEGER start;
LARGE_INTEGER stop;


char* myLoadProgSource(const char* cFilename, const char* cPreamble, size_t* szFinalLength)
{
    // locals 
    FILE* pFileStream = NULL;
    size_t szSourceLength;

    pFileStream = fopen(cFilename, "rb");
    if(pFileStream == 0) 
    {       
        return NULL;
    }

    size_t szPreambleLength = strlen(cPreamble);

    // get the length of the source code
    fseek(pFileStream, 0, SEEK_END); 
    szSourceLength = ftell(pFileStream);
    fseek(pFileStream, 0, SEEK_SET); 

    // allocate a buffer for the source code string and read it in
    char* cSourceString = (char *)malloc(szSourceLength + szPreambleLength + 1); 
    memcpy(cSourceString, cPreamble, szPreambleLength);
    if (fread((cSourceString) + szPreambleLength, szSourceLength, 1, pFileStream) != 1)
    {
        fclose(pFileStream);
        free(cSourceString);
        return 0;
    }

    // close the file and return the total length of the combined (preamble + source) string
    fclose(pFileStream);
    if(szFinalLength != 0)
    {
        *szFinalLength = szSourceLength + szPreambleLength;
    }
    cSourceString[szSourceLength + szPreambleLength] = '\0';

    return cSourceString;
}

void addition(type* input1, float input2, type* output, unsigned int elements){
	for(unsigned int i=0;i<elements;i++){
		output[i] = input1[i] + input2;
	}
}


int main(int argc, char** const argv)
{

	size_t elements = DATA_SIZE;
	size_t d = 0;
	//std::cin >> elements;
	//OpenCL variables
	cl_context context;
	cl_context_properties properties[3];
	
	cl_command_queue command_queue;
	cl_program program;
	cl_int err;
	cl_uint num_of_platforms;
	cl_platform_id platform_id = NULL;
	cl_uint num_of_devices = 0;
	
	cl_kernel addKernel;
	
	cl_mem clInput1;
	cl_mem clInput2;
	
	cl_mem clAddOutput;
	
	
	
	cl_event add_event;
	

	size_t global;
	size_t local;
	
	float nwork_groups;
	size_t work_group_size;
	size_t work_item_size[3];

	
	

	for(d=0; d<2; d++){
		for(elements=1024; elements<=33554432; elements*=2){

			printf("elements: %lu ",elements);
			printf("device: %lu ",d);

			QueryPerformanceFrequency(&frequency);
			
			//GPU input and output arrays
			type* inputData1 = new type[elements];
			type* inputData2 = new type[elements];
			
			type* add_output = new type[elements];
			type* add_output_gpu = new type[elements];

			srand((unsigned)time(0));
	
			//populating input with random values

			for(int sz=0; sz<elements; sz++)
			{
				inputData1[sz] = float(rand()) + (float)((rand() << 15 + rand()) & ((1 << 24) - 1)) / (1 << 24);
				inputData2[sz] = float(rand()) + (float)((rand() << 15 + rand()) & ((1 << 24) - 1)) / (1 << 24);
				
			}
			////printf("\n");
	
			/*for(int i=0;i<elements;i++){
				printf("%f ",inputData1[i]);
			}*/

			//calling cpu version of select
			

			QueryPerformanceCounter(&start);
			for(int i=0;i<1;i++){
				addition(inputData1,1,add_output, elements);
				
			}
			QueryPerformanceCounter(&stop);

			double timespent = (double)(stop.QuadPart - start.QuadPart) / (double)frequency.QuadPart;
			printf ("Microbenchmark E on CPU took total without looping = %f seconds\n", timespent/1);

	
			//(1) retreive the first available platform
			if (clGetPlatformIDs(1, &platform_id, &num_of_platforms) != CL_SUCCESS)
			{
				printf("Unable to get platform_id\n");
				return 1;
			}
			//printf("Platform number: %d\n", num_of_platforms);

			//get platform name
			char tmpBuffer[1024];
			if ((err = clGetPlatformInfo(platform_id, CL_PLATFORM_NAME, sizeof(tmpBuffer), tmpBuffer, NULL)) != CL_SUCCESS)
			{
				printf("%d, Unable to get platform name!\n", err);
				return 1;
			}
			//printf("Platform name: %s\n", tmpBuffer);

			//get platform version
			if (clGetPlatformInfo(platform_id, CL_PLATFORM_VERSION, sizeof(tmpBuffer), tmpBuffer, NULL) != CL_SUCCESS)
			{
				printf("Unable to get platform version!\n");
				return 1;
			}
			//printf("Platform version: %s\n", tmpBuffer);
 
			//(2) try to get GPU device(s)

			// Get number of devices available 
			//cl_uint deviceount = 0;
			if(err = clGetDeviceIDs(platform_id, CL_DEVICE_TYPE_ALL, 0, NULL, &num_of_devices) != CL_SUCCESS)
			{
				printf("Unable to get device_id, %d\n", err);
				return 1;
			}

			cl_device_id* device_id = (cl_device_id*)malloc(sizeof(cl_device_id) * num_of_devices);
	
			if(num_of_devices == 0)
			{
				printf("No devices found!\n");
				return 1;
			}
			//printf("Number of devices: %d\n", num_of_devices);

			if ((err = clGetDeviceIDs(platform_id, CL_DEVICE_TYPE_ALL, num_of_devices, device_id, NULL)) != CL_SUCCESS)
			{
				printf("clGetDeviceIDs failed, %d\n", err);
				return 1;
			}
	
			//get device name

			if (clGetDeviceInfo(device_id[d], CL_DEVICE_NAME, sizeof(tmpBuffer), &tmpBuffer, NULL) != CL_SUCCESS)
			{
				printf("Unable to get device name\n");
				return 1;
			} 
			printf("Device %u Name: %s\n", 0, tmpBuffer);


			//get max compute units in device
			cl_uint comp_unit;   
			err = clGetDeviceInfo(device_id[d], CL_DEVICE_MAX_COMPUTE_UNITS, sizeof(cl_uint), &comp_unit, NULL);
			if (err != CL_SUCCESS)
			{
	        		printf("Error: Failed to access device number of compute units !\n");
	        		return 1;
			}
			printf("with a max of %d compute units \n",comp_unit);

			//get local memory size on device
			cl_ulong tmpLong;
			if(clGetDeviceInfo(device_id[d], CL_DEVICE_LOCAL_MEM_SIZE, sizeof(tmpLong), &tmpLong, NULL) != CL_SUCCESS) {
				printf("Unable to get local memory size\n");
				return 1;
			}
			//printf("Device local mem size: %ld\n", tmpLong);
			//printf("Input size: %lu ",sizeof(type)*elements);
		
			//get max memory allocation on device
			if(clGetDeviceInfo(device_id[d],  CL_DEVICE_MAX_MEM_ALLOC_SIZE , sizeof(tmpLong), &tmpLong, NULL) != CL_SUCCESS) {
				printf("Unable to get max memory size\n");
				return 1;
			}
			//printf("Device memory allocation size: %ld\n", tmpLong);
			//printf("Buffer size allocated %ld\n", sizeof(type) * elements);

			//get max no. of work items in a work group on a single compute unit
			err = clGetDeviceInfo( device_id[d], CL_DEVICE_MAX_WORK_GROUP_SIZE, sizeof(size_t), &work_group_size, NULL); 
			if (err != CL_SUCCESS)
			{
				printf("Error: Failed to get device Info (CL_DEVICE_MAX_WORK_GROUP_SIZE)!\n",err);
				return 1;
			}
			//printf("max work group size: %lu\n", work_group_size);

			//get Maximum number of work-items that can be specified in each dimension of the work-group to clEnqueueNDRangeKernel
			err = clGetDeviceInfo( device_id[d], CL_DEVICE_MAX_WORK_ITEM_SIZES, sizeof(size_t)*3, work_item_size, NULL); 
			if (err != CL_SUCCESS)
			{
				printf("Error: Failed to get device Info (CL_DEVICE_MAX_WORK_ITEM_SIZES)!\n",err);
				return 1;
			}
			//for(int i=0;i<3;i++)
				//printf("max work item size: %lu ", work_item_size[i]);
			//printf("\n");


			//(3) context properties list - must be terminated with 0
			properties[0]= CL_CONTEXT_PLATFORM;
			properties[1]= (cl_context_properties) platform_id;
			properties[2]= 0;

			// create a context with the GPU device
			context = clCreateContext(properties,1,&device_id[d],NULL,NULL,&err);
			//printf("context, %x, err %d, \n", context, err);

			//(4) create command queue using the context and device
			command_queue = clCreateCommandQueue(context, device_id[d], CL_QUEUE_PROFILING_ENABLE, &err);
			//printf("command queue, %x, err %d, \n", command_queue, err);
	
			//(5) Get Source code
			char * ProgramSource = myLoadProgSource("test.cl", "", NULL);
			if(!ProgramSource)
			{
				printf("Failed to load program source!\n");
				system("pause");
				return 1;	
			}

			//(6) create a program from the kernel source code
			program = clCreateProgramWithSource(context,1,(const char **) &ProgramSource, NULL, &err);
			//printf("program, %x, err %d, \n", program, err);

			//(7) compile the program
			if ((err = clBuildProgram(program, 1, &device_id[d], NULL, NULL, NULL)) != CL_SUCCESS)
			{
				printf("Error building program, err %d\n", err);

				char *build_log;
				size_t ret_val_size;
				err = clGetProgramBuildInfo(program, device_id[d], CL_PROGRAM_BUILD_LOG, 0, NULL, &ret_val_size);
				
				build_log = (char*)malloc(ret_val_size+1);
				err = clGetProgramBuildInfo(program, device_id[d], CL_PROGRAM_BUILD_LOG, ret_val_size, build_log, NULL);
       
				build_log[ret_val_size] = '\0';
				fprintf(stderr, "%s", build_log );
				system("pause");
				return 1;
			}
	
			//(8) specify which kernel from the program to execute
			addKernel = clCreateKernel(program, "addition", &err);
			//printf("return kernel %lx, err %d\n", kernel, err);
			
			


			local = work_group_size;  
			global = 7*local*comp_unit; //7 wavefronts
			if(global < elements){
				while( (elements) % global != 0 )
					global += local;
			}
			else
				global = elements/4; 
	
			printf("global:%d ",global);

			unsigned int ctas = global/local;
		


/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
			//(9) create buffers for the input and ouput
			clInput1 = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_MEM_USE_HOST_PTR, sizeof(type) * elements, inputData1, &err);
			//printf("return input %lx, err %d\n", clInput, err);
			clInput2 = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_MEM_USE_HOST_PTR, sizeof(type) * elements, inputData2, &err);
			//printf("return input %lx, err %d\n", clInput, err);
			
			clAddOutput = clCreateBuffer(context, CL_MEM_READ_WRITE, sizeof(type) * elements, NULL, &err);
			//printf("return input %lx, err %d\n", clInput, err);
			


	//		err = clEnqueueWriteBuffer(command_queue, clInput, CL_TRUE, 0, sizeof(type)*elements, inputData, 0, NULL, NULL);

			/*QueryPerformanceCounter(&start);
			for(int i=0;i<NLOOPS;i++){
				err = clEnqueueWriteBuffer(command_queue, clInput, CL_TRUE, 0, sizeof(type)*elements, inputData, 0, NULL, NULL);
			}
			QueryPerformanceCounter(&stop);
			double time = (double)(stop.QuadPart - start.QuadPart) / (double)frequency.QuadPart;
			printf ("writing input took total = %f seconds\n", time/NLOOPS);
			*/
			cl_ulong map_time_start = 0, map_time_end = 0, unmap_time_start = 0, unmap_time_end = 0, page_map_time_start = 0, page_map_time_end = 0;
			double map_time=0,unmap_time=0,map_time_taken=0,unmap_time_taken=0,page_map_time=0,page_map_time_taken=0;
			void* input_ptr;
			void* pages_ptr;

			/*QueryPerformanceCounter(&start);

			for(int i=0;i<NLOOPS;i++){
				input_ptr = clEnqueueMapBuffer (command_queue, clInput, CL_TRUE, CL_MAP_WRITE_INVALIDATE_REGION, 0, sizeof(type) * elements, 0, NULL, &map_event, &err);
				clGetEventProfilingInfo(map_event, CL_PROFILING_COMMAND_START, sizeof(map_time_start), &map_time_start, NULL);
				clGetEventProfilingInfo(map_event, CL_PROFILING_COMMAND_END, sizeof(map_time_end), &map_time_end, NULL);
				map_time_taken = map_time_end - map_time_start;
				map_time +=map_time_taken;

				memcpy(input_ptr,inputData, sizeof(type) * elements);
				clEnqueueUnmapMemObject (command_queue, clInput , input_ptr, 0,  NULL, &unmap_event);
				
			}
			
			QueryPerformanceCounter(&stop);

			printf("Map event time (mapping input):%f s\n",(map_time/1e9)/NLOOPS);
			double time = (double)(stop.QuadPart - start.QuadPart) / (double)frequency.QuadPart;
			printf ("Mapping and copying input took total = %f seconds\n", time/NLOOPS);*/
			
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

			

			err = clSetKernelArg(addKernel, 0, sizeof(void*), (void*)&clInput1);
			//printf("Set arg err %d\n", err);
			err = clSetKernelArg(addKernel, 1, sizeof(void*), (void*)&clAddOutput);
			//printf("Set arg err %d\n", err);
			err = clSetKernelArg(addKernel, 2, sizeof(int), &elements);
			//printf("Set arg err %d\n", err);

			

			int nloops = NLOOPS;

			double atotal_time=0,stotal_time=0,m1total_time=0,mtotal_time=0;
			map_time = 0;
			type *output_ptr;


			QueryPerformanceCounter(&start);
			while(nloops--)
			{

				cl_ulong atime_start=0, atime_end=0;
				double atime_taken=0;

				// enqueue the kernel command for execution
				err = clEnqueueNDRangeKernel(command_queue, addKernel, 1, NULL, &global, &local, 0, NULL, &add_event); 
				////printf("launch kernel err %d\n", err);
				clFinish(command_queue);
	
				

				
				
				// copy the results from out of the buffer
				err = clEnqueueReadBuffer(command_queue, clAddOutput, CL_TRUE, 0, sizeof(type)*elements, add_output_gpu, 0, NULL, NULL);
				//printf("reading buffer err %d\n", err);
				

				//output_ptr = (type*) clEnqueueMapBuffer (command_queue, clOutput, CL_TRUE, CL_MAP_READ, 0, sizeof(type) * output_size, 0, NULL, &map_event, &err);

				clGetEventProfilingInfo(add_event, CL_PROFILING_COMMAND_START, sizeof(atime_start), &atime_start, NULL);
				clGetEventProfilingInfo(add_event, CL_PROFILING_COMMAND_END, sizeof(atime_end), &atime_end, NULL);
				atime_taken = atime_end - atime_start;
				atotal_time += atime_taken;

				

				/*clGetEventProfilingInfo(map_event, CL_PROFILING_COMMAND_START, sizeof(map_time_start), &map_time_start, NULL);
				clGetEventProfilingInfo(map_event, CL_PROFILING_COMMAND_END, sizeof(map_time_end), &map_time_end, NULL);
				map_time_taken = map_time_end - map_time_start;
				map_time +=map_time_taken;*/
			}
  			QueryPerformanceCounter(&stop);

			//printf("Map event time (reading output):%f s\n",(map_time/1e9)/NLOOPS);
			printf("\n");
			atotal_time = atotal_time/NLOOPS;
			printf("GPU add took total time (event profiling)= %f seconds\n", (atotal_time / 1e9) );
			

			timespent = (double)(stop.QuadPart - start.QuadPart) / (double)frequency.QuadPart;
			printf ("Select on GPU took total = %f seconds\n", timespent/NLOOPS);

			

			// print the results
			int c = 0;
			for(int i=0; i<elements; i++)
			{
				
				if(add_output_gpu[i] != add_output[i])
					c++;
				
	
			}
			printf("Difference count:%d ",c);
			printf("\n\n");
			
			
	
			delete [] inputData1;
			delete [] inputData2;
			
			delete [] add_output;
			
	}
}
	
	// cleanup - release OpenCL resources
	clReleaseMemObject(clInput1);
	clReleaseMemObject(clInput2);
	
	clReleaseMemObject(clAddOutput);
	
	clReleaseProgram(program);
	clReleaseKernel(addKernel);
	
	clReleaseCommandQueue(command_queue);
	clReleaseContext(context);

	system("pause");
	return 0;
}



